/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: BinaryFunctionals.h,v 1.15 2003/11/30 18:35:17 opetzold Exp $
 */

#ifndef TVMET_BINARY_FUNCTIONAL_H
#define TVMET_BINARY_FUNCTIONAL_H

namespace tvmet {


/**
 * \class Fcnl_Assign BinaryFunctionals.h "tvmet/BinaryFunctionals.h"
 * \brief Binary operator for assign operations.
 *
 * Unfortunally we have sometimes to cast on assign operations e.g.,
 * on assign on different POD. So we avoid warnings.
 */
template <class T1, class T2>
struct Fcnl_Assign : public BinaryFunctional {
  typedef void						value_type;
  static inline value_type apply_on(T1& _tvmet_restrict lhs, T2 rhs) {
    lhs = static_cast<T1>(rhs);
  }
  static void print_xpr(std::ostream& os, std::size_t l=0) {
    os << IndentLevel(l) << "fncl_Assign<T1="
       << typeid(T1).name() << ", T2=" << typeid(T2).name() << ">,"
       << std::endl;
  }
};


/** \class Fcnl_AddUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_SubUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_MulUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_DivUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_ModUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseXorUpd	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseAndUpd	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseOrUpd	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_ShlUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_ShrUpd 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
#define TVMET_IMPLEMENT_MACRO(OPNAME, OP)                   		\
template <class T1, class T2>                                 		\
struct OPNAME : public BinaryFunctional {	          		\
  typedef void						value_type; 	\
  static inline value_type apply_on(T1& _tvmet_restrict lhs, T2 rhs) { 	\
    lhs OP rhs;                                           		\
  }                                                         		\
  static void print_xpr(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #OPNAME << "<T1="  				\
       << typeid(T1).name() << ", T2=" << typeid(T2).name() << ">," 	\
       << std::endl;  							\
  }     								\
};

TVMET_IMPLEMENT_MACRO(Fcnl_AddUpd, +=)
TVMET_IMPLEMENT_MACRO(Fcnl_SubUpd, -=)
TVMET_IMPLEMENT_MACRO(Fcnl_MulUpd, *=)
TVMET_IMPLEMENT_MACRO(Fcnl_DivUpd, /=)
TVMET_IMPLEMENT_MACRO(Fcnl_ModUpd, %=)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXorUpd, ^=)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAndUpd, &=)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOrUpd, |=)
TVMET_IMPLEMENT_MACRO(Fcnl_ShlUpd, <<=)
TVMET_IMPLEMENT_MACRO(Fcnl_ShrUpd, >>=)

#undef TVMET_IMPLEMENT_MACRO


/** \class Fcnl_Add 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Sub 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Mul 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Div 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Mod 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseXor	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseAnd	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_BitwiseOr	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Shl 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Shr 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
#define TVMET_IMPLEMENT_MACRO(OPNAME, OP)                   		\
template <class T1, class T2>                                 		\
struct OPNAME : public BinaryFunctional {              			\
  typedef typename  PromoteTraits<T1, T2>::value_type	value_type; 	\
  static inline value_type apply_on(T1 lhs, T2 rhs) {          		\
    return lhs OP rhs;                                     		\
  }                                                         		\
  static void print_xpr(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #OPNAME << "<T1="  				\
       << typeid(T1).name() << ", T2=" << typeid(T2).name() << ">,"    	\
       << std::endl;  							\
  }     								\
};

TVMET_IMPLEMENT_MACRO(Fcnl_Add, +)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, -)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, *)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, /)
TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |)
TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<)
TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>)

#undef TVMET_IMPLEMENT_MACRO


/** \class Fcnl_Greater 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_GreaterOrEqual 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Less 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_LessOrEqual 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_Equal 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_NotEqual 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_LogicalAnd 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_LogicalOr 	BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
#define TVMET_IMPLEMENT_MACRO(OPNAME, OP)                   		\
template <class T1, class T2>                            		\
struct OPNAME : public BinaryFunctional {          			\
  typedef bool						value_type; 	\
  static inline value_type apply_on(T1 lhs, T2 rhs) {          		\
    return lhs OP rhs;                                     		\
  }                                                         		\
  static void print_xpr(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #OPNAME << "<T1="  				\
       << typeid(T1).name() << ", T2=" << typeid(T2).name() << ">,"    	\
       << std::endl;  							\
  }     								\
};

TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO


/** \class Fcnl_atan2 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_drem 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_fmod 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_hypot 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_jn 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_yn 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
/** \class Fcnl_pow 		BinaryFunctionals.h "tvmet/BinaryFunctionals.h" */
#define TVMET_IMPLEMENT_MACRO(FUNCNAME, MATHFUNC)             		\
template <class T1, class T2>	                               		\
struct FUNCNAME : public BinaryFunctional {	              		\
  typedef typename PromoteTraits<T1, T2>::value_type	value_type; 	\
  static inline value_type apply_on(T1 lhs, T2 rhs) {	        	\
    return MATHFUNC(lhs, rhs);						\
  }                                                         		\
  static void print_xpr(std::ostream& os, std::size_t l=0) {    		\
    os << IndentLevel(l) << #FUNCNAME << "<T1="				\
       << typeid(T1).name() << ", T2=" << typeid(T2).name() << ">,"    	\
       << std::endl;  							\
  }     								\
};

TVMET_IMPLEMENT_MACRO(Fcnl_atan2, std::atan2)
TVMET_IMPLEMENT_MACRO(Fcnl_drem, std::drem)
TVMET_IMPLEMENT_MACRO(Fcnl_fmod, std::fmod)
TVMET_IMPLEMENT_MACRO(Fcnl_hypot, std::hypot)
TVMET_IMPLEMENT_MACRO(Fcnl_jn, std::jn)
TVMET_IMPLEMENT_MACRO(Fcnl_yn, std::yn)
TVMET_IMPLEMENT_MACRO(Fcnl_pow, std::pow)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/**
 * \class Fcnl_polar BinaryFunctionals.h "tvmet/BinaryFunctionals.h"
 * \brief %Functional for polar.
 */
template <class T1, class T2> struct Fcnl_polar : public BinaryFunctional { };


/**
 * \class Fcnl_polar<T,T> BinaryFunctionals.h "tvmet/BinaryFunctionals.h"
 * \brief %Functional for polar.
 * \note  This functional is partialy specialized due to the declaration
 *        of %polar in namespace std <tt>complex<T> polar(T, T)</tt>.
 *        This means especially that type promotion isn't avaible here.
 */
template <class T>
struct Fcnl_polar<T,T> : public BinaryFunctional {
  typedef std::complex<T>                               value_type;
  static inline value_type apply_on(T lhs, T rhs) {
    return std::polar(lhs, rhs);
  }
  static void print_xpr(std::ostream& os, std::size_t l=0) {
    os << IndentLevel(l) << "Fcnl_polar<T1="
       << typeid(T).name() << ", T2=" << typeid(T).name() << ">,"
       << std::endl;
  }
};
#endif // defined(TVMET_HAVE_COMPLEX)


/**
 * \class Fcnl_swap BinaryFunctionals.h "tvmet/BinaryFunctionals.h"
 * \brief Binary operator for swapping values using temporaries.
 * \todo check for std::swap implementation; todo before LUdecomp
 */
template <class T1, class T2>
struct Fcnl_swap : public BinaryFunctional {
  typedef void						value_type;
  static inline value_type apply_on(T1& _tvmet_restrict lhs, T2& _tvmet_restrict rhs) {
    typedef typename  PromoteTraits<T1, T2>::value_type	temp_type;

    temp_type 						temp(lhs);
    lhs = static_cast<T1>(rhs);
    rhs = static_cast<T2>(tmp);
  }
  static void print_xpr(std::ostream& os, std::size_t l=0) {
    os << IndentLevel(l) << "Fcnl_swap<T1="
       << typeid(T1).name() << ", T2" << typeid(T2).name() << ">,"
       << std::endl;
  }
};


} // namespace tvmet

#endif // TVMET_BINARY_FUNCTIONAL_H

// Local Variables:
// mode:C++
// End:
