/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: VectorOperators.h,v 1.13 2003/11/30 08:26:25 opetzold Exp $
 */

#ifndef TVMET_VECTOR_OPERATORS_H
#define TVMET_VECTOR_OPERATORS_H

namespace tvmet {


/**
 * \fn operator<<(std::ostream& os, const Vector<T, Sz>& rhs)
 * \brief Overload operator for i/o
 * \ingroup _binary_operator
 */
template<class T, std::size_t Sz>
inline
std::ostream& operator<<(std::ostream& os, const Vector<T, Sz>& rhs) {
  return rhs.print_on(os);
}


/***********************************************************************
 * Member operators (arithmetic and bit ops)
 ***********************************************************************/


/*
 * update_operator(Vector<T1, Sz>,  Vector<T2, Sz>)
 * update_operator(Vector<T1, Sz>,  XprVector<E, Sz>)
 * Note: per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(MEMBER, OP)	                   	\
template<class T1, class T2, std::size_t Sz>				\
inline Vector<T1, Sz>& 							\
operator OP (Vector<T1, Sz>& lhs, const Vector<T2, Sz>& rhs) {		\
  return lhs.MEMBER(rhs);						\
}									\
									\
template<class T, class E, std::size_t Sz>				\
inline Vector<T, Sz>& 							\
operator OP (Vector<T, Sz>& lhs, const XprVector<E, Sz>& rhs) {		\
  return lhs.MEMBER(rhs);						\
}

TVMET_IMPLEMENT_MACRO(AddUpd, +=)		// per se element wise
TVMET_IMPLEMENT_MACRO(SubUpd, -=)		// per se element wise
TVMET_IMPLEMENT_MACRO(MulUpd, *=)		// per se element wise

namespace element_wise {
  TVMET_IMPLEMENT_MACRO(DivUpd, /=)		// not defined for vectors
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(ModUpd, %=)
  TVMET_IMPLEMENT_MACRO(BitwiseXorUpd, ^=)
  TVMET_IMPLEMENT_MACRO(BitwiseAndUpd, &=)
  TVMET_IMPLEMENT_MACRO(BitwiseOrUpd, |=)
  TVMET_IMPLEMENT_MACRO(ShlUpd, <<=)
  TVMET_IMPLEMENT_MACRO(ShrUpd, >>=)
}

#undef TVMET_IMPLEMENT_MACRO


/************************************************************************
 * Vector arithmetic operators implemented by functions
 * add, sub, mul and div
 ************************************************************************/


/*
 * operator(Vector<T1, Sz>, Vector<T2, Sz>)
 * operator(Vector<T1, Sz>, XprVector<E, Sz>)
 * operator(XprVector<E, Sz>, Vector<T1, Sz>)
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)                   	\
template<class T1, class T2, std::size_t Sz> 	                  	\
inline                 			      			      	\
XprVector<                   						\
  XprBinOp<                   						\
    FUNCTIONAL<T1, T2>,                   				\
    VectorConstReference<T1, Sz>,              				\
    VectorConstReference<T2, Sz>               				\
  >,									\
  Sz									\
>                   			              			\
operator OP (const Vector<T1, Sz>& lhs, const Vector<T2, Sz>& rhs) { 	\
  return FUNC (lhs, rhs);						\
}                      	      						\
                      	      						\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprVector<E, Sz>,                  	      				\
    VectorConstReference<T, Sz>                	      			\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (const XprVector<E, Sz>& lhs, const Vector<T, Sz>& rhs) {	\
  return FUNC (lhs, rhs);						\
}                    	      						\
                    	      						\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    VectorConstReference<T, Sz>,               	      			\
    XprVector<E, Sz>                   	      				\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (const Vector<T, Sz>& lhs, const XprVector<E, Sz>& rhs) {  	\
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)		// per se element wise

namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)	// not defined for vectors
}

#undef TVMET_IMPLEMENT_MACRO


/*
 * operator(Vector<T, Sz>, POD)
 * operator(POD, Vector<T, Sz>)
 * Note: operations +,-,*,/ are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP, POD)               	\
template<class T, std::size_t Sz>   					\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< T, POD >,          					\
    VectorConstReference<T, Sz>,      					\
    XprLiteral< POD >          						\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<T, Sz>& lhs, POD rhs) {			\
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class T, std::size_t Sz>	 				\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< POD, T>,          					\
    XprLiteral< POD >,          					\
    VectorConstReference<T, Sz>       					\
  >,          								\
  Sz									\
>          								\
operator OP (POD lhs, const Vector<T, Sz>& rhs) {         		\
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, int)

#if defined(TVMET_HAVE_LONG_LONG)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long long int)
#endif

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, float)

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long double)
#endif

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(Vector<std::complex<T>, Sz>, std::complex<T>)
 * operator(std::complex<T>, Vector<std::complex<T>, Sz>)
 * Note: operations +,-,*,/ are per se element wise
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)	               	\
template<class T, std::size_t Sz>              				\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    VectorConstReference< std::complex<T>, Sz>,				\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<std::complex<T>, Sz>& lhs, const std::complex<T>& rhs) { \
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class T, std::size_t Sz>     					\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >,          				\
    VectorConstReference< std::complex<T>, Sz>				\
  >,          								\
  Sz									\
>          								\
operator OP (const std::complex<T>& lhs, const Vector< std::complex<T>, Sz>& rhs) { \
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)		// per se element wise
#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/************************************************************************
 * Vector integer and compare operators
 ************************************************************************/


/*
 * operator(Vector<T1, Sz>, Vector<T2, Sz>)
 * operator(XprVector<E, Sz>, Vector<T, Sz>)
 * operator(Vector<T, Sz>, XprVector<E, Sz>)
 * Note: operations are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                   	\
template<class T1, class T2, std::size_t Sz> 	                  	\
inline                 			      			      	\
XprVector<                   						\
  XprBinOp<                   						\
    FUNCTIONAL<T1, T2>,                   				\
    VectorConstReference<T1, Sz>,              				\
    VectorConstReference<T2, Sz>               				\
  >,									\
  Sz									\
>                   			              			\
operator OP (const Vector<T1, Sz>& lhs, const Vector<T2, Sz>& rhs) { 	\
  typedef XprBinOp <                    				\
    FUNCTIONAL<T1, T2>,  	                 			\
    VectorConstReference<T1, Sz>,              				\
    VectorConstReference<T2, Sz>                     	      		\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(lhs.const_ref(), rhs.const_ref()));			\
}                   			              			\
                   			              			\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprVector<E, Sz>,                  	      				\
    VectorConstReference<T, Sz>                	      			\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (const XprVector<E, Sz>& lhs, const Vector<T, Sz>& rhs) {	\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprVector<E, Sz>,                  	      				\
    VectorConstReference<T, Sz>                	      			\
  > 							 expr_type; 	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(lhs, rhs.const_ref()));					\
}                    	      						\
                    	      						\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    VectorConstReference<T, Sz>,               	      			\
    XprVector<E, Sz>                   	      				\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (const Vector<T, Sz>& lhs, const XprVector<E, Sz>& rhs) {  	\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    VectorConstReference<T, Sz>,      	      				\
    XprVector<E, Sz>                  	      				\
  > 						 	expr_type; 	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(lhs.const_ref(), rhs));         				\
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO



#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(Vector<std::complex<T>, Sz>, std::complex<T>)
 * operator(std::complex<T>, Vector<std::complex<T>, Sz>)
 * Note: - per se element wise
 *       - bit ops on complex<int> doesn't make sense, stay away
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)		               	\
template<class T, std::size_t Sz>              				\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    VectorConstReference< std::complex<T>, Sz>,				\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<std::complex<T>, Sz>& lhs, const std::complex<T>& rhs) { \
  typedef XprBinOp<          						\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    VectorConstReference< std::complex<T>, Sz>,				\
    XprLiteral< std::complex<T> >          				\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(lhs.const_ref(), XprLiteral< std::complex<T> >(rhs))); 	\
}          								\
          								\
template<class T, std::size_t Sz>     					\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >,          				\
    VectorConstReference< std::complex<T>, Sz>				\
  >,          								\
  Sz									\
>          								\
operator OP (const std::complex<T>& lhs, const Vector< std::complex<T>, Sz>& rhs) { \
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, std::complex<T> >,			\
    XprLiteral< std::complex<T> >,         				\
    VectorConstReference< std::complex<T>, Sz>				\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(XprLiteral< std::complex<T> >(lhs), rhs.const_ref())); 	\
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/*
 * operator(Vector<T, Sz>, POD)
 * operator(POD, Vector<T, Sz>)
 * Note: operations are per se element_wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP, TP)               	\
template<class T, std::size_t Sz>              				\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< T, TP >,          					\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< TP >          						\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<T, Sz>& lhs, TP rhs) {			\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, TP >,          					\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< TP >          						\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(lhs.const_ref(), XprLiteral< TP >(rhs))); 		\
}          								\
          								\
template<class T, std::size_t Sz>     					\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< TP, T>,          					\
    XprLiteral< TP >,          						\
    VectorConstReference<T, Sz>       					\
  >,          								\
  Sz									\
>          								\
operator OP (TP lhs, const Vector<T, Sz>& rhs) {         		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, T>,         					\
    XprLiteral< TP >,         						\
    VectorConstReference<T, Sz>       					\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(                 	      		\
    expr_type(XprLiteral< TP >(lhs), rhs.const_ref())); 		\
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, int)

#if defined(TVMET_HAVE_LONG_LONG)
// integer operators only
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, long long int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, float)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, float)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, float)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


/***********************************************************************
 * global unary operators
 ***********************************************************************/


/*
 * unary_operator(Vector<T, Sz>)
 * Note: per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                   	\
template <class T, std::size_t Sz>                                      \
inline                                                        		\
XprVector<                                                    		\
  XprUnOp<                                                        	\
    FUNCTIONAL<T>,                                                      \
    VectorConstReference<T, Sz>                                       	\
  >,                                                        		\
  Sz									\
>                                                        		\
operator OP (const Vector<T, Sz>& rhs) {                                \
  typedef XprUnOp<                                                      \
    FUNCTIONAL<T>,                                                      \
    VectorConstReference<T, Sz>                                        	\
  >  							 expr_type;	\
  return XprVector<expr_type, Sz>(expr_type(rhs.const_ref()));          \
}

TVMET_IMPLEMENT_MACRO(Fcnl_LogicalNot, !)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseNot, ~)
TVMET_IMPLEMENT_MACRO(Fcnl_Negate, -)
#undef TVMET_IMPLEMENT_MACRO


} // namespace tvmet

#endif // TVMET_VECTOR_OPERATORS_H

// Local Variables:
// mode:C++
// End:
