/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MatrixOperators.h,v 1.16 2003/11/30 08:26:25 opetzold Exp $
 */

#ifndef TVMET_XPR_MATRIX_OPERATORS_H
#define TVMET_XPR_MATRIX_OPERATORS_H

namespace tvmet {


/************************************************************************
 * Matrix arithmetic operators implemented by functions
 * add, sub, mul and div
 ************************************************************************/


/*
 * operator(const XprMatrix<E1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1,Cols2>& rhs)
 * Note: operations +,-,*,/ are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)                   	\
template<class E1, std::size_t Rows1, std::size_t Cols1,            	\
         class E2, std::size_t Cols2>					\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows1, Cols1>,       			      		\
    XprMatrix<E2, Cols2, Cols1>        			      		\
  >,                   			      			      	\
  Rows1, Cols2           		      			      	\
>                   			      		      		\
operator OP (const XprMatrix<E1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1, Cols2>& rhs) { \
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)		// per se element wise
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)		// per se element wise

namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)	// see as prod()
  TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)	// not defined for matrizes, must be element_wise
}
#undef TVMET_IMPLEMENT_MACRO


/*
 * operator(XprMatrix<E, Rows, Cols>,  POD)
 * operator(POD, XprMatrix<E, Rows, Cols>)
 * Note: operations +,-,*,/ are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP, POD)               	\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, POD >,      			\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< POD >          						\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, POD rhs) {		\
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class E,std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<POD, typename E::value_type>,    			\
    XprLiteral< POD >,          					\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (POD lhs, const XprMatrix<E, Rows, Cols>& rhs) {  		\
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, int)

#if defined(TVMET_HAVE_LONG_LONG)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, float)

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(XprMatrix<E, Rows, Cols>, complex<>)
 * operator(complex<>, XprMatrix<E, Rows, Cols>)
 * Note: operations +,-,*,/ are per se element wise
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, FUNC, OP)            		\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,      		\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, const std::complex<T>& rhs) { \
  return FUNC (lhs, rhs);						\
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<std::complex<T>, typename E::value_type>,    		\
    XprLiteral< std::complex<T> >,          				\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const std::complex<T>& lhs, const XprMatrix<E, Rows, Cols>& rhs) { \
  return FUNC (lhs, rhs);						\
}

TVMET_IMPLEMENT_MACRO(Fcnl_Add, add, +)
TVMET_IMPLEMENT_MACRO(Fcnl_Sub, sub, -)
TVMET_IMPLEMENT_MACRO(Fcnl_Mul, mul, *)
TVMET_IMPLEMENT_MACRO(Fcnl_Div, div, /)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/************************************************************************
 * matrix specific operator*() = prod() operations
 ************************************************************************/


/**
 * \fn operator*(const XprMatrix<E1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1, Cols2>& rhs)
 * \brief Evaluate the product of two XprMatrix.
 * \ingroup _binary_operator
 * \sa prod(XprMatrix<E1, Rows1, Cols1> lhs, XprMatrix<E2, Cols1, Cols2> rhs)
 */
template<class E1, std::size_t Rows1, std::size_t Cols1,
	 class E2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    XprMatrix<E1, Rows1, Cols1>, Rows1, Cols1,	// M1(Rows1, Cols1)
    XprMatrix<E2, Cols1, Cols2>, Cols2,		// M2(Cols1, Cols2)
    Cols1, 1,					// Stride M1
    Cols2, 1					// Stride M2
  >,
  Rows1, Cols2
>
operator*(const XprMatrix<E1, Rows1, Cols1>& lhs, const XprMatrix<E2, Cols1, Cols2>& rhs) {
  return prod(lhs, rhs);
}


/************************************************************************
 * matrix-vector specific prod( ... ) operators
 ************************************************************************/

/**
 * \fn operator*(const XprMatrix<E1, Rows, Cols>& lhs, const XprVector<E2, Cols>& rhs)
 * \brief Evaluate the product of XprMatrix and XprVector.
 * \ingroup _binary_operator
 * \sa prod(XprMatrix<E1, Rows, Cols> lhs, XprVector<E2, Cols> rhs)
 */
template<class E1, std::size_t Rows, std::size_t Cols,
	 class E2>
inline
XprVector<
  XprMVProduct<
    XprMatrix<E1, Rows, Cols>, Rows, Cols,
    Cols, 1,
    XprVector<E2, Cols>, 1
  >,
  Rows
>
operator*(const XprMatrix<E1, Rows, Cols>& lhs, const XprVector<E2, Cols>& rhs) {
  return prod(lhs, rhs);
}


/************************************************************************
 * Matrix integer and compare operators
 ************************************************************************/


/*
 * operator(XprMatrix<>, XprMatrix<>)
 * Note: operations are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                   	\
template<class E1, std::size_t Rows, std::size_t Cols,            	\
         class E2>							\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,       			      		\
    XprMatrix<E2, Rows, Cols>        			      		\
  >,                   			      			      	\
  Rows, Cols	           		      			      	\
>                   			      		      		\
operator OP (const XprMatrix<E1, Rows, Cols>& lhs, const XprMatrix<E2, Rows, Cols>& rhs) { \
  typedef XprBinOp<							\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprMatrix<E1, Rows, Cols>,						\
    XprMatrix<E2, Rows, Cols>						\
  >		    					expr_type; 	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(lhs, rhs));		\
}

// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * operator(XprMatrix<E, Rows, Cols>, std::complex<>)
 * operator(std::complex<>, XprMatrix<E, Rows, Cols>)
 * Note: - per se element wise
 *       - bit ops on complex<int> doesn't make sense, stay away
 * \todo type promotion
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP, TP)              		\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,      		\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, const std::complex<T>& rhs) { \
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, std::complex<T> >,   		\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< std::complex<T> >              			      	\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(     			      	\
    expr_type(lhs, XprLiteral< std::complex<T> >(rhs))); 		\
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols, class T>		\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<std::complex<T>, typename E::value_type>,    		\
    XprLiteral< std::complex<T> >,          				\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const std::complex<T>& lhs, const XprMatrix<E, Rows, Cols>& rhs) { \
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, typename E::value_type>,   		\
    XprLiteral< std::complex<T> >,         				\
    XprMatrix<E, Rows, Cols>   			      			\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(     			      	\
    expr_type(XprLiteral< std::complex<T> >(lhs), rhs)); 		\
}


// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, std::complex<T>)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, std::complex<T>)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


/*
 * operator(XprMatrix<E, Rows, Cols>, POD)
 * operator(POD, XprMatrix<E, Rows, Cols>)
 * Note: operations are per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP, TP)              		\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,      			\
    XprMatrix<E, Rows, Cols>,		          			\
    XprLiteral< TP >          						\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (const XprMatrix<E, Rows, Cols>& lhs, TP rhs) {		\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >,   				\
    XprMatrix<E, Rows, Cols>,		        			\
    XprLiteral< TP >              			      		\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(     			      	\
    expr_type(lhs, XprLiteral< TP >(rhs))); 				\
}          								\
          								\
template<class E, std::size_t Rows, std::size_t Cols>			\
inline                 			      			      	\
XprMatrix<          							\
  XprBinOp<          							\
    FUNCTIONAL<TP, typename E::value_type>,    				\
    XprLiteral< TP >,          						\
    XprMatrix<E, Rows, Cols>			      			\
  >,          								\
  Rows, Cols   								\
>          								\
operator OP (TP lhs, const XprMatrix<E, Rows, Cols>& rhs) {  		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, typename E::value_type>,   				\
    XprLiteral< TP >,         						\
    XprMatrix<E, Rows, Cols>   			      			\
  >							expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(     			      	\
    expr_type(XprLiteral< TP >(lhs), rhs)); 				\
}


// integer operators only, e.g used on double you wil get an error
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, int)

#if defined(TVMET_HAVE_LONG_LONG)
// integer operators only
namespace element_wise {
  TVMET_IMPLEMENT_MACRO(Fcnl_Mod, %, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseXor, ^, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseAnd, &, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseOr, |, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shl, <<, long long int)
  TVMET_IMPLEMENT_MACRO(Fcnl_Shr, >>, long long int)
}

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long long int)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, float)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, float)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, float)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, float)

// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
// necessary operators for eval functions
TVMET_IMPLEMENT_MACRO(Fcnl_Greater, >, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Less, <, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_GreaterOrEqual, >=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LessOrEqual, <=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_Equal, ==, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_NotEqual, !=, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalAnd, &&, long double)
TVMET_IMPLEMENT_MACRO(Fcnl_LogicalOr, ||, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


/***********************************************************************
 * global unary operators
 ***********************************************************************/


/*
 * unary_operator(const XprMatrix<E, Rows, Cols>& m)
 * Note: per se element wise
 */
#define TVMET_IMPLEMENT_MACRO(FUNCTIONAL, OP)                   	\
template <class E, std::size_t Rows, std::size_t Cols>			\
inline                                                        		\
XprMatrix<                                                        	\
  XprUnOp<                                                        	\
    FUNCTIONAL<typename E::value_type>,                                 \
    XprMatrix<E, Rows, Cols>						\
  >,                                                        		\
  Rows, Cols								\
>                                                        		\
operator OP (const XprMatrix<E, Rows, Cols>& m) {			\
  typedef XprUnOp<                                                      \
    FUNCTIONAL<typename E::value_type>,                                 \
    XprMatrix<E, Rows, Cols>						\
  >  							 expr_type;	\
  return XprMatrix<expr_type, Rows, Cols>(expr_type(m));		\
}

TVMET_IMPLEMENT_MACRO(Fcnl_LogicalNot, !)
TVMET_IMPLEMENT_MACRO(Fcnl_BitwiseNot, ~)
TVMET_IMPLEMENT_MACRO(Fcnl_Negate, -)

#undef TVMET_IMPLEMENT_MACRO


} // namespace tvmet

#endif // TVMET_XPR_MATRIX_OPERATORS_H

// Local Variables:
// mode:C++
// End:
